#!/usr/bin/env python

from __future__ import print_function

import os, sys

from util import use_qt5
if use_qt5 :
   from PyQt5.QtCore import *
   from PyQt5.QtGui import *
   from PyQt5.QtWidgets import *
else :
   from PyQt4.QtCore import *
   from PyQt4.QtGui import *

from lexer     import Lexer
from grammar   import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal
from symbols   import initSymbols
from recognize import initRecognize, findRule, recognizeRule
from tohtml    import htmlFromRules
from tolout    import loutFromRules
from toparser  import initCollections, parserFromRules, printCollections, selectKeywords
from toproduct import productFromRules

# --------------------------------------------------------------------------

# class ViewWindow
#    variables
#       tree : QTreeWidget
#    methods
#       getEditor () : Editor
#       hasExtension (editor: Editor, extension:string) : boolean

# class Editor
#    variables
#       origFileName
#       edit: QTextEdit

# main function
#    QIcon.setThemeName ("oxygen")

# --------------------------------------------------------------------------

def findIcon (icon_name) :
    return QIcon.fromTheme (icon_name)

class TreeItem (QTreeWidgetItem) :
    def __init__ (self, parent, text) :
        super (TreeItem, self).__init__ (parent)
        self.setText (0, text)

    def addIcon (self, icon_name) :
        self.setIcon (0, findIcon (icon_name))

# --------------------------------------------------------------------------

class GrammarPlugin (QObject) :

   def __init__ (self, main_window) :
       super (GrammarPlugin, self).__init__ (main_window)
       self.win = main_window

       self.gramMenu = self.win.menuBar().addMenu ("&Grammar")
       self.gramMenu.aboutToShow.connect (self.onShowGramMenu)

       act = QAction ("&Compile", self.win)
       act.setShortcut ("F9")
       act.triggered.connect (self.compileFile)
       self.gramMenu.addAction (act)

       act = QAction ("to &HTML", self.win)
       act.triggered.connect (self.toHtml)
       self.gramMenu.addAction (act)

       act = QAction ("to &Lout", self.win)
       act.triggered.connect (self.toLout)
       self.gramMenu.addAction (act)

       act = QAction ("to &Parser", self.win)
       act.setShortcut ("F11")
       act.triggered.connect (self.toParser)
       self.gramMenu.addAction (act)

       act = QAction ("to P&roduct", self.win)
       act.setShortcut ("F12")
       act.triggered.connect (self.toProduct)
       self.gramMenu.addAction (act)

       act = QAction ("Recognize", self.win)
       act.triggered.connect (self.recognize)
       self.gramMenu.addAction (act)

   def onShowGramMenu (self) :
       is_gram = self.win.hasExtension (self.win.getEditor(), ".g")
       for act in self.gramMenu.actions () :
          act.setEnabled (is_gram)

   def onItemActivated (self, item, column) :
       if hasattr (item, "line") :
          self.edit.selectLine (item.line)

   def outputFile (self, fileName, extension) :
       fileName = os.path.basename (fileName)
       fileName = os.path.join ("_output", fileName)
       fileName, ext = os.path.splitext (fileName)
       fileName = fileName + extension
       return fileName

   def createDir (self, fileName) :
       dirName = os.path.dirname (fileName)
       if not os.path.isdir (dirName) :
          os.makedirs (dirName)

   # Html

   def toHtml (self) :
       e = self.win.getEditor ()
       if self.win.hasExtension (e, ".g") :
          source = str (e.edit.toPlainText ())

          fileName = self.outputFile (e.origFileName, ".html")
          self.createDir (fileName)

          save_file = sys.stdout
          sys.stdout = open (fileName, "w")

          grammar = Grammar ()
          grammar.openString (source)

          grammar.parseRules ()
          htmlFromRules (grammar)

          sys.stdout = save_file

          self.win.reloadFile (fileName)
          self.win.showHtml (fileName)
          print ("O.K.")

   # Lout

   def toLout (self) :
       e = self.win.getEditor ()
       if self.win.hasExtension (e, ".g") :
          source = str (e.edit.toPlainText ())

          fileName = self.outputFile (e.origFileName, ".lout")
          self.createDir (fileName)

          save_file = sys.stdout
          sys.stdout = open (fileName, "w")

          grammar = Grammar ()
          grammar.openString (source)

          grammar.parseRules ()
          loutFromRules (grammar)

          sys.stdout = save_file
          self.win.reloadFile (fileName)

          # create PDF or PostScript file

          subdir = os.path.dirname (fileName)
          save_dir = os.getcwd ()
          os.chdir (subdir)

          fileName = os.path.basename (fileName) # local file name

          toPdf = False

          if toPdf :

             pdfFileName, ext = os.path.splitext (fileName)
             pdfFileName = pdfFileName + ".pdf"

             os.system ("lout" + " " +  fileName + " -PDF -o " + pdfFileName)
             self.win.showPdf (pdfFileName)

          else :

             psFileName, ext = os.path.splitext (fileName)
             psFileName = psFileName + ".ps"

             # yum install lout
             os.system ("lout" + " " +  fileName + " -o " + psFileName)
             # lout also creates .li file in current directory

             self.win.showPdf (psFileName)

          os.chdir (save_dir)
          print ("O.K.")

   # To Parser

   def toParser (self) :
       e = self.win.getEditor ()
       if self.win.hasExtension (e, ".g") :
          source = str (e.edit.toPlainText ())

          fileName = self.outputFile (e.origFileName, "-parser.py")
          self.createDir (fileName)

          save_file = sys.stdout
          sys.stdout = open (fileName, "w")

          grammar = Grammar ()
          grammar.openString (source)

          grammar.parseRules ()
          initSymbols (grammar)

          initCollections (grammar)
          parserFromRules (grammar)
          printCollections (grammar)
          selectKeywords (grammar)

          sys.stdout = save_file
          print ("O.K.")
          self.win.reloadFile (fileName)

   # To Product

   def toProduct (self) :
       e = self.win.getEditor ()
       if self.win.hasExtension (e, ".g") :
          source = str (e.edit.toPlainText ())

          fileName = self.outputFile (e.origFileName, "-product.py")
          self.createDir (fileName)

          save_file = sys.stdout
          sys.stdout = open (fileName, "w")

          grammar = Grammar ()
          grammar.openString (source)

          grammar.parseRules ()
          initSymbols (grammar)

          productFromRules (grammar)

          sys.stdout = save_file
          print ("O.K.")
          self.win.reloadFile (fileName)

   # Recognize

   def recognize (self) :
       e = self.win.getEditor ()
       if self.win.hasExtension (e, ".g") :
          source = str (e.edit.toPlainText ())

          grammar = Grammar ()
          grammar.openString (source)

          grammar.parseRules ()
          initSymbols (grammar)

          # !?
          lexer = Lexer ()
          lexer.pascal = True
          initRecognize (lexer, grammar)
          lexer.openFile ("../grm/idedemo.pas") # !?
          rule = findRule (grammar, "module_decl")
          recognizeRule (lexer, grammar, rule)

   # Compile Grammar

   def compileFile (self) :
       e = self.win.getEditor ()
       if self.win.hasExtension (e, ".g") :
          source = str (e.edit.toPlainText ())

          grammar = Grammar ()
          grammar.openString (source)
          # print source

          grammar.parseRules ()
          initSymbols (grammar)

          print ("parsed")

          branch = TreeItem (self.win.tree, "grammar " + os.path.basename (e.origFileName))
          branch.fileName = e.origFileName

          self.grammar = grammar
          self.addSymbols (branch)
          for rule in grammar.rules :
             self.addBranch (branch, rule)

          print ("O.K.")

   def addSymbols (self, above) :
       branch = TreeItem (above, "symbols")
       for symbol in self.grammar.symbols :
           TreeItem (branch, str (symbol.inx) + " " + symbol.name)

   def addBranch (self, above, node) :

       txt = ""
       if isinstance (node, Rule) :
          txt = node.name
       elif isinstance (node, Expression) :
          txt = "expression"
       elif isinstance (node, Alternative) :
          txt = "alternative"
       elif isinstance (node, Ebnf) :
          txt = "ebnf " + node.mark
       elif isinstance (node, Nonterminal) :
          txt = "nonterminal "
          if node.variable != "" :
             txt = txt + node.variable + ":"
          txt = txt + node.rule
          # if node.symbol != None :
          #    txt = txt + " (symbol " + node.symbol.name + ")"
       elif isinstance (node, Terminal) :
          txt = "terminal " + node.text
          # if node.symbol != None :
          #    txt = txt + " (symbol " + node.symbol.name + ")"
       else:
          txt = node.__class__.__name__

       item = TreeItem (above, txt)

       if hasattr (node, "line") :
          item.line = node.line

       if isinstance (node, Rule) :
          item.setToolTip (0, "line " + str (node.line))
          item.addIcon ("code-class")
       if isinstance (node, Nonterminal) :
          item.addIcon ("code-function")
       if isinstance (node, Terminal) :
          item.addIcon ("code-variable")
       if isinstance (node, Ebnf) :
          item.addIcon ("code-block")

       if isinstance (node, Rule) :
          if hasattr (node, "first") :
             n = 0
             for inx in range (len (node.first)) :
                 if node.first [inx] :
                    n = n + 1
             if n > 1 :
                item.setForeground (0, QBrush (Qt.green))

       if hasattr (node, "nullable") :
          if node.nullable :
             item.setForeground (0, QBrush (Qt.red))

       self.addInfo (item, node)

       if isinstance (node, Rule) :
          # self.addInfo (item, node)
          self.addBranch (item, node.expr)
       elif isinstance (node, Expression) :
          for t in node.alternatives :
             self.addBranch (item, t)
       elif isinstance (node, Alternative) :
          for t in node.items :
             self.addBranch (item, t)
       elif isinstance (node, Ebnf) :
          self.addBranch (item, node.expr)
          if node.impl != None :
             self.addBranch (item, node.impl)

   def addInfo (self, above, rule) :
       if hasattr (rule, "first") :
          branch = TreeItem (above, "first")
          branch.addIcon ("info")
          branch.setForeground (0, QBrush (Qt.blue))
          for inx in range (len (rule.first)) :
              if rule.first [inx] :
                 name = self.grammar.symbols [inx].name
                 item = TreeItem (branch, name)
                 item.setForeground (0, QBrush (Qt.blue))

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
