#!/usr/bin/env python

from __future__ import print_function

from lexer import Lexer, Separators
from grammar import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal, Directive
from info import incIndent, decIndent, put, putEol, putCondEol, putLn
from symbols import Symbol, initSymbols, initLexerKeywords, initLexerSeparators

# --------------------------------------------------------------------------

class TargetEnvironment (object) :
   def __init__ (self) :
      self.class_dict = { }
      self.super_dict = { }
      self.dynamic_globals = globals ()
      self.dynamic_locals = locals ()

def initRecognize (lexer, grammar) :
    initLexerKeywords (lexer, grammar)
    initLexerSeparators (lexer, grammar)

    env = TargetEnvironment ()

    for rule in grammar.rules :
       if rule.super_type != "" :
          env.super_dict [rule.rule_type] = rule.super_type

    for rule in grammar.rules :
        createClass (rule.rule_type, env)

    return env

# --------------------------------------------------------------------------

def createClass (rule_type, env) :
    if rule_type != "" :
       if rule_type not in env.class_dict :
          env.class_dict [rule_type] = True

          super_type = ""
          if rule_type in env.super_dict :
             super_type = env.super_dict [rule_type]
          if super_type != "" and super_type not in env.class_dict :
             createClass (super_type, env)
          if super_type == "" :
             super_type = "object"

          eol = "\n"
          tab = "\t"

          code = "class " + rule_type
          if super_type != "" :
             code = code + " (" + super_type + ")"
          code = code + " :" + eol
          code = code + tab + "pass" + eol
          print (code)
          exec (code, env.dynamic_globals, env.dynamic_locals)

def createRuleObject (rule, env) :
    rule_type = rule.rule_type
    if rule_type == "" :
       rule_type = "object"

    eol = "\n"
    tab = "\t"

    code = rule_type + " ()"
    print (code)
    target = eval (code, env.dynamic_globals, env.dynamic_locals)
    return target

def storeNonterminal (nonterm, env, target, value) :
    print ("field " + target.__class__.__name__ + " . " + nonterm.variable)
    if nonterm.variable != "" :
       print ("set field " + nonterm.variable + " = " + str (value))
       setattr (target, nonterm.variable, value)
    elif nonterm.add :
       if not hasattr (target, "items") :
          target.items = [ ]
       print ("add item " + str (value))
       target.items.append (value)

# --------------------------------------------------------------------------

def symbolIn (lexer, first) :
    return first [lexer.token]

def findRule (grammar, name) :
    rule = None
    for r in grammar.rules :
        if r.name  == name :
           rule = r
           break
    if rule == None :
       grammar.error ("Unknown rule " + name)
    return rule

# --------------------------------------------------------------------------

def recognizeRule (lexer, grammar, rule, env) :
    putLn ("rule " + rule.name)
    incIndent ()
    target = createRuleObject (rule, env)
    recognizeExpression (lexer, grammar, rule.expr, env, target)
    decIndent ()
    putLn ("end of rule " + rule.name)
    return target

def recognizeExpression (lexer, grammar, expr, env, target) :
    n = 0
    found = False
    for alt in expr.alternatives :
        n = n + 1
        # putLn ("is it alternative " + str (n))
        if alt.nullable or symbolIn (lexer, alt.first) :
            putLn ("alternative " + str (n))
            recognizeAlternative (lexer, grammar, alt, env, target)
            found = True
            break
    if not found :
       lexer.error ("Unknown alternative, on input" + grammar.symbols [lexer.token].name + ", " + lexer.tokenText)

def recognizeAlternative (lexer, grammar, alt, env, target) :
    for item in alt.items :
        if isinstance (item, Terminal) :
           recognizeTerminal (lexer, grammar, item, env, target)
        elif isinstance (item, Nonterminal) :
           recognizeNonterminal (lexer, grammar, item, env, target)
        elif isinstance (item, Ebnf) :
           recognizeEbnf (lexer, grammar, item, env, target)
        elif isinstance (item, Directive) :
           pass
        else :
           raise Exception ("Unknown alternative item")

def recognizeTerminal (lexer, grammar, item, env, target) :
    if lexer.token == item.symbol.inx :
       putLn ("terminal " + item.text)
       lexer.nextToken ()
    else :
       lexer.error (item.text + " expected")

def recognizeNonterminal (lexer, grammar, item, env, target) :
    value = None
    if item.symbol != None :
       if lexer.token != item.symbol.inx :
          lexer.error (item.symbol.name.capitalize() + " expected")
       putLn (item.symbol.name + " " + lexer.tokenText)
       value = lexer.tokenText
       lexer.nextToken ()
    else :
       rule = findRule (grammar, item.rule)
       value = recognizeRule (lexer, grammar, rule, env)
    storeNonterminal (item, env, target, value)

def recognizeEbnf (lexer, grammar, item, env, target) :
    if item.mark == "*" :
       while symbolIn (lexer, item.first) :
          recognizeExpression (lexer, grammar, item.expr, env, target)
    elif item.mark == "?" :
       if symbolIn (lexer, item.first) :
          recognizeExpression (lexer, grammar, item.expr, env, target)
    elif item.mark == "+" :
       recognizeExpression (lexer, grammar, item.expr, env, target)
       while symbolIn (lexer, item.first) :
          recognizeExpression (lexer, grammar, item.expr, env, target)
    else:
       recognizeExpression (lexer, grammar, item.expr, env, target)

# --------------------------------------------------------------------------

def printStructure (target) :
    if isinstance (target, str) :
       putLn ("string " + target)
    else :
       putLn ("object " + target.__class__.__name__)
       incIndent ()
       if hasattr (target, "__dict__") : # !?
          for ident in target.__dict__ :
             if ident != "items" :
                item = target.__dict__ [ident]
                putLn ("field " + ident)
                incIndent ()
                printStructure (item)
                decIndent ()

       if hasattr (target, "items") :
          putLn ("items:")
          incIndent ()
          for item in target.items :
             printStructure (item)
          decIndent ()
       decIndent ()

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    grammar = Grammar ()
    grammar.openFile ("pas.g")
    grammar.parseRules ()
    initSymbols (grammar)

    lexer = Lexer ()
    lexer.pascal = True
    env = initRecognize (lexer, grammar)
    lexer.openFile ("idedemo.pas")
    # rule = findRule (grammar, "module_decl")
    rule = findRule (grammar, "unit_decl")
    target = recognizeRule (lexer, grammar, rule, env)
    printStructure (target)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
