#!/usr/bin/env python

from __future__ import print_function

from lexer import Separators
from grammar import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal, Directive
from info import incIndent, decIndent, put, putEol, putCondEol, putLn

# --------------------------------------------------------------------------

class Symbol (object) :
   def __init__ (self) :
       self.inx = 0
       self.name = ""

# --------------------------------------------------------------------------

def addToSeparators (tree, symbol, pos) :
    c = symbol.name [pos]
    if c not in tree :
       tree[c] = Separators ()
    if pos == len (symbol.name) - 1 :
       tree[c].value = symbol.inx
    else :
       addToSeparators (tree[c].items, symbol, pos + 1)

def printSeparators (tree, grammar) :
    for c in tree :
       putLn ("\"" + c + "\" :")
       branch = tree [c]
       incIndent ()
       if branch.value >= 0 :
          symbol =  grammar.symbols [branch.value]
          putLn ("separator " + str (branch.value) + " " + symbol.name)
       printSeparators (branch.items, grammar)
       decIndent ()

def printLexerSeparators (lexer, grammar) :
    printSeparators (lexer.lexer_separators, grammar)

def printLexerKeywords (lexer, grammar) :
    for keyword in sorted (lexer.lexer_keywords) :
        k = lexer.lexer_keywords [keyword]
        symbol = grammar.symbols [k]
        putLn (str (k) + " " + symbol.name) #  " " + keyword

def initLexerKeywords (lexer, grammar) :
    lexer.lexer_keywords = { }
    for k in grammar.keywords :
        symbol = grammar.keywords [k]
        lexer.lexer_keywords [symbol.name] = symbol.inx
    # printLexerKeywords (lexer, grammar)

def initLexerSeparators (lexer, grammar) :
    lexer.lexer_separators = { }
    for k in grammar.delimiters :
        symbol = grammar.delimiters [k]
        addToSeparators (lexer.lexer_separators, symbol, 0)
    # printLexerSeparators (lexer, grammar)

# --------------------------------------------------------------------------

def addSymbol (grammar, symbol) :
    symbol.inx = grammar.symbol_cnt
    grammar.symbol_cnt = grammar.symbol_cnt + 1
    grammar.symbols.append (symbol)

def symbolsFromRules (grammar) :

    grammar.keywords = { }
    grammar.delimiters = { }
    grammar.special_list = [ ]
    grammar.special_dict = {  }

    for name in ["identifier", "number", "real_number", "character_literal", "string_literal", "unknown_separator", "end_of_line"] :
        symbol = Symbol ()
        symbol.name = "<" + name + ">"
        grammar.special_list.append (symbol)
        grammar.special_dict [name] = symbol

    for rule in grammar.rules :
        symbolsFromExpression (grammar, rule.expr)

    grammar.symbol_cnt = 0
    grammar.symbols = [ ]

    empty = Symbol ()
    empty.name = "empty";
    addSymbol (grammar, empty)

    for symbol in grammar.special_list :
        addSymbol (grammar, symbol)

    for key in sorted (grammar.keywords.keys()) :
        symbol = grammar.keywords [key]
        addSymbol (grammar, symbol)

    for key in sorted (grammar.delimiters.keys()) :
        symbol = grammar.delimiters [key]
        addSymbol (grammar, symbol)

    # for symbol in grammar.symbols :
    #     print symbol.inx, symbol.name

# --------------------------------------------------------------------------

def symbolsFromExpression (grammar, expr) :
    for alt in expr.alternatives :
       symbolsFromAlternative (grammar, alt)

def symbolsFromAlternative (grammar, alt) :
    for item in alt.items :
        if isinstance (item, Terminal) :
           symbolsFromTerminal (grammar, item)
        elif isinstance (item, Nonterminal) :
           symbolsFromNonterminal (grammar, item)
        elif isinstance (item, Ebnf) :
           symbolsFromEbnf (grammar, item)
        elif isinstance (item, Directive) :
            pass
        else :
           raise Exception ("Unknown alternative item " + item.__class__.__name__)

def symbolsFromEbnf (grammar, ebnf) :
    symbolsFromExpression (grammar, ebnf.expr)
    if ebnf.impl != None :
       symbolsFromAlternative (grammar, ebnf.impl)

def symbolsFromNonterminal (grammar, item) :
    name = item.rule
    if name in grammar.special_dict.keys() :
        item.symbol = grammar.special_dict [name]
        # print "PRIMITIVE ", item.symbol.inx, item.symbol.name
    else :
        item.symbol = None
        # print "NONTERMINAL", name

def symbolsFromTerminal (grammar, item) :
    name = item.text
    if grammar.isLetter (name [0]) :
       if name in grammar.keywords :
          symbol = grammar.keywords [name]
       else :
          symbol = Symbol ()
          symbol.name = name
          grammar.keywords [name] = symbol
    else :
       if name in grammar.delimiters :
          symbol = grammar.delimiters [name]
       else :
          symbol = Symbol ()
          symbol.name = name
          grammar.delimiters [name] = symbol
    item.symbol = symbol
    # print "TERMINAL", name

# --------------------------------------------------------------------------

def nullableRules (grammar) :
    grammar.rule_dict = { }

    for rule in grammar.rules :
        rule.nullable = False

        if rule.name in grammar.rule_dict :
           raise Exception ("Rule " + rule.name + " already defined")
        grammar.rule_dict [rule.name] = rule

    grammar.nullableChanged = True
    while grammar.nullableChanged :
       grammar.nullableChanged = False
       # print "nullable step"
       for rule in grammar.rules :
           nullableRule (grammar, rule)

def nullableRule (grammar, rule) :
    expr = rule.expr
    nullableExpression (grammar, expr)
    if rule.nullable != expr.nullable :
       rule.nullable = expr.nullable
       grammar.nullableChanged = True

def nullableExpression (grammar, expr) :
    # init
    expr.nullable = (len (expr.alternatives) == 0)

    for alt in expr.alternatives :
       nullableAlternative (grammar, alt)

       # one alternative is nullable => expression is nullable
       if alt.nullable :
          expr.nullable = True

def nullableAlternative (grammar, alt) :
    # init
    alt.nullable = True

    for item in alt.items :
        if isinstance (item, Terminal) :
           nullableTerminal (grammar, item)
        elif isinstance (item, Nonterminal) :
           nullableNonterminal (grammar, item)
        elif isinstance (item, Ebnf) :
           nullableEbnf (grammar, item)
        elif isinstance (item, Directive) :
           pass
        else :
           raise Exception ("Unknown alternative item: " + item.__class__.__name__)

        if not isinstance (item, Directive) :
           # one item is not nullable => alternative is also not nullable
           if not item.nullable :
              alt.nullable = False

def nullableEbnf (grammar, ebnf) :
    nullableExpression (grammar, ebnf.expr)
    if ebnf.impl != None :
       nullableAlternative (grammar, ebnf.impl)

    expr = ebnf.expr
    # set ebnf according to expression
    ebnf.nullable = expr.nullable
    # ( )? or ( )* => ebnf is nullable
    if ebnf.mark == '?' or ebnf.mark == '*' :
       ebnf.nullable = True

def nullableNonterminal (grammar, item) :
    if item.symbol != None :
       # primitive non-terminal (identifier, number, string literal)
       item.nullable = False
    else :
       # calculate nullable for non-terminal
       name = item.rule
       if name not in grammar.rule_dict:
           raise Exception ("Unknown rule: " + name)
       rule = grammar.rule_dict [name]
       item.nullable = rule.nullable

def nullableTerminal (grammar, item) :
    item.nullable = False

# --------------------------------------------------------------------------

def firstFromRules (grammar) :
    grammar.rule_dict = { }

    for rule in grammar.rules :
        rule.first = newArray (grammar)

        if rule.name in grammar.rule_dict :
           raise Exception ("Rule " + rule.name + " already defined")
        grammar.rule_dict [rule.name] = rule

    # calculate first sets
    grammar.firstChanged = True
    while grammar.firstChanged :
       grammar.firstChanged = False
       # print "first step"
       for rule in grammar.rules :
           firstFromRule (grammar, rule)

def firstFromRule (grammar, rule) :
    expr = rule.expr
    firstFromExpression (grammar, expr)

    equal = True
    for inx in range (grammar.symbol_cnt) :
        if rule.first [inx] != expr.first [inx] :
           equal = False

    if not equal :
       rule.first = expr.first
       grammar.firstChanged = True

def newArray (grammar) :
    return [False] * grammar.symbol_cnt

def firstFromExpression (grammar, expr) :
    # init
    expr.first = newArray (grammar)

    for alt in expr.alternatives :
       firstFromAlternative (grammar, alt)

       # add symbols from alternative to expression
       for inx in range (grammar.symbol_cnt) :
           if alt.first [inx] :
              if expr.first [inx] :
                 pass # raise Exception ("Conflict between alternatives")
              expr.first [inx] = True

def firstFromAlternative (grammar, alt) :
    # init
    alt.first = newArray (grammar)
    add = True

    for item in alt.items :
        if isinstance (item, Terminal) :
           firstFromTerminal (grammar, item)
        elif isinstance (item, Nonterminal) :
           firstFromNonterminal (grammar, item)
        elif isinstance (item, Ebnf) :
           firstFromEbnf (grammar, item)
        elif isinstance (item, Directive) :
           pass
        else :
           raise Exception ("Unknown alternative item: " + item.__class__.__name__)

        if not isinstance (item, Directive) :
           if add :
              for inx in range (grammar.symbol_cnt) :
                  if item.first [inx] :
                     alt.first [inx] = True
           # item is not nullable => stop adding symbols from items
           if not item.nullable :
              add = False

def firstFromEbnf (grammar, ebnf) :
    firstFromExpression (grammar, ebnf.expr)
    if ebnf.impl != None :
       firstFromAlternative (grammar, ebnf.impl)

    expr = ebnf.expr
    # set ebnf according to expression
    ebnf.first = expr.first

def firstFromNonterminal (grammar, item) :
    if item.symbol != None :
       # primitive non-terminal (identifier, number, string literal)
       item.first = newArray (grammar)
       inx = item.symbol.inx
       item.first [inx] = True
    else :
       # calculate first set for non-terminal
       name = item.rule
       if name not in grammar.rule_dict:
           raise Exception ("Unknown rule " + name)
       rule = grammar.rule_dict [name]
       item.first = rule.first

def firstFromTerminal (grammar, item) :
    # first set has one item
    item.first = newArray (grammar)
    inx = item.symbol.inx
    item.first [inx] = True

# --------------------------------------------------------------------------

def initSymbols (grammar) :
    symbolsFromRules (grammar)
    nullableRules (grammar)
    firstFromRules (grammar)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
