#!/usr/bin/env python

import sys, os

from util import use_qt5
if use_qt5 :
   from PyQt5.QtCore import *
   from PyQt5.QtGui import *
   from PyQt5.QtWidgets import *
else :
   from PyQt4.QtCore import *
   from PyQt4.QtGui import *

sys.path.insert (1, os.path.join (os.getcwd (), "../grm")) # directory with parser

from lexer   import Lexer
from grammar import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal
from symbols import symbolsFromRules, firstFromRules

# --------------------------------------------------------------------------

class Highlighter (QSyntaxHighlighter) :
   def __init__ (self, parent = None) :
       super (Highlighter, self).__init__ (parent)
       self.lexer = Lexer ()

       self.keywordFormat = QTextCharFormat ()
       self.keywordFormat.setForeground (Qt.darkRed)

       self.identifierFormat = QTextCharFormat ()
       self.identifierFormat.setForeground (QColor (0, 0, 0))

       self.qidentifierFormat = QTextCharFormat ()
       self.qidentifierFormat.setForeground (Qt.green)

       self.numberFormat = QTextCharFormat ()
       self.numberFormat.setForeground (Qt.blue)

       self.realFormat = QTextCharFormat ()
       self.realFormat.setForeground (Qt.magenta)

       self.characterFormat = QTextCharFormat ()
       self.characterFormat.setForeground (QColor ("cornflowerblue"))

       self.stringFormat = QTextCharFormat ()
       self.stringFormat.setForeground (Qt.blue)

       self.separatorFormat = QTextCharFormat ()
       self.separatorFormat.setForeground (QColor ("orange"))

       self.commentFormat = QTextCharFormat ()
       self.commentFormat.setForeground (Qt.gray)

   def highlightBlock (self, text) :
       try :
          lexer = self.lexer
          lexer.openString (text)
          while not lexer.isEndOfSource () :
             index = lexer.tokenByteOfs
             length = lexer.charByteOfs - lexer.tokenByteOfs
             if lexer.token == lexer.identifier :
                if lexer.tokenText in ["class", "def"] :
                   self.setFormat (index, length, self.keywordFormat)
                elif lexer.tokenText[0] == 'Q' :
                   self.setFormat (index, length, self.qidentifierFormat)
                else :
                   self.setFormat (index, length, self.identifierFormat)
             elif lexer.token == lexer.number :
                self.setFormat (index, length, self.numberFormat)
             elif lexer.token == lexer.real :
                self.setFormat (index, length, self.realFormat)
             elif lexer.token == lexer.character :
                self.setFormat (index, length, self.characterFormat)
             elif lexer.token == lexer.string :
                self.setFormat (index, length, self.stringFormat)
             elif lexer.token == lexer.separator :
                self.setFormat (index, length, self.separatorFormat)
             else : # !?
                self.setFormat (index, length, self.commentFormat)
             lexer.nextToken ()
       except :
          pass

# --------------------------------------------------------------------------

class CompletionTextEdit (QTextEdit) :
   def __init__ (self, parent=None) :
       super (CompletionTextEdit, self).__init__ (parent)
       # self.setMinimumWidth (400)
       self.completer = None
       # self.moveCursor (QTextCursor.End)

   def setCompleter (self, completer, unfiltered = False) :
       if self.completer :
           # self.disconnect (self.completer, SIGNAL ("activated(const QString&)"),  self.insertCompletion)
           self.completer.activated.disconnect (self.insertCompletion)
       if not completer :
           return

       completer.setWidget (self)
       if unfiltered :
          completer.setCompletionMode (QCompleter.UnfilteredPopupCompletion)
       else :
          completer.setCompletionMode (QCompleter.PopupCompletion)
       completer.setCaseSensitivity (Qt.CaseInsensitive)
       self.completer = completer
       # self.connect (self.completer, SIGNAL ("activated(const QString&)"), self.insertCompletion)
       self.completer.activated.connect (self.insertCompletion)

   def insertCompletion (self, completion) :
       tc = self.textCursor ()
       extra = (completion.length() -  self.completer.completionPrefix().length())
       tc.movePosition (QTextCursor.Left)
       tc.movePosition (QTextCursor.EndOfWord)
       tc.insertText (completion.right(extra))
       self.setTextCursor (tc)

   def textUnderCursor (self) :
       tc = self.textCursor ()
       tc.select (QTextCursor.WordUnderCursor)
       return tc.selectedText ()

   def focusInEvent (self, event) :
       if self.completer :
           self.completer.setWidget (self);
       QTextEdit.focusInEvent (self, event)

   def keyPressEvent (self, event) :
       if self.completer and self.completer.popup().isVisible() :
           if event.key () in (
              Qt.Key_Enter,
              Qt.Key_Return,
              Qt.Key_Escape,
              Qt.Key_Tab,
              Qt.Key_Backtab) :
                 event.ignore ()
                 return

       ## has Ctrl+Space been pressed??
       isShortcut = (event.modifiers () == Qt.ControlModifier and
                     (event.key () == Qt.Key_Space))
       if (not self.completer or not isShortcut) :
           QTextEdit.keyPressEvent (self, event)

       ## ctrl or shift key on it's own??
       ctrlOrShift = event.modifiers () in (Qt.ControlModifier,  Qt.ShiftModifier)
       if ctrlOrShift and event.text ().isEmpty () :
           # ctrl or shift key on it's own
           return

       eow = QString ("~!@#$%^&*()_+{}|:\"<>?,./;'[]\\-=") #end of word

       hasModifier = ((event.modifiers () != Qt.NoModifier) and not ctrlOrShift)

       completionPrefix = self.textUnderCursor ()

       if (not isShortcut and (hasModifier or event.text ().isEmpty () or
       completionPrefix.length () < 3 or
       eow.contains (event.text().right(1)))) :
           self.completer.popup ().hide ()
           return

       if (completionPrefix != self.completer.completionPrefix ()) :
           self.completer.setCompletionPrefix (completionPrefix)
           popup = self.completer.popup ()
           popup.setCurrentIndex (self.completer.completionModel ().index (0,0))

       cr = self.cursorRect ()
       cr.setWidth (self.completer.popup ().sizeHintForColumn (0)
           + self.completer.popup ().verticalScrollBar ().sizeHint ().width ())
       self.completer.complete (cr) ## popup it up!

# http://rowinggolfer.blogspot.cz/2010/08/qtextedit-with-autocompletion-using.html

# --------------------------------------------------------------------------

class IndentationTextEdit (CompletionTextEdit) :

   def __init__ (self, parent=None) :
       super (IndentationTextEdit, self).__init__ (parent)

   def indent (self) :
       delta = 1
       cursor = self.textCursor ()
       if cursor.hasSelection () :
           start = cursor.selectionStart ()
           stop = cursor.selectionEnd ()

           cursor.setPosition (start)
           start_line = cursor.blockNumber ()

           # Move to the selection's last line
           cursor.setPosition (stop)
           stop_line = cursor.blockNumber ()

           for n in range (stop_line - start_line + 1) :
               # Move to start of each line
               cursor.movePosition (QTextCursor.StartOfLine)
               # Insert tabbing
               cursor.insertText (" " * delta)
               # And move back up
               cursor.movePosition (QTextCursor.Up)
       else :
           # If there is no selection, just insert a tab
           cursor.movePosition (QTextCursor.StartOfLine)
           cursor.insertText (" " * delta)

   def unindent (self) :
       cursor = self.textCursor ()
       if cursor.hasSelection () :
           start = cursor.selectionStart ()
           stop = cursor.selectionEnd ()

           cursor.setPosition (start)
           start_line = cursor.blockNumber ()

           # Move to the selection's last line
           cursor.setPosition (stop)
           stop_line = cursor.blockNumber ()

           # Iterate over lines
           for n in range (stop_line - start_line + 1) :
               self.unindentLine (cursor)
               cursor.movePosition (QTextCursor.Up)
       else :
           self.unindentLine (cursor)

   def unindentLine (self, cursor) :
       delta = 1
       cursor.movePosition (QTextCursor.StartOfLine)

       # Grab the current line
       line = cursor.block().text ()

       # If the line starts with a tab character, delete it
       if line.startsWith ("\t") :
           # Delete next character
           cursor.deleteChar ()
           cursor.insertText (" " * (delta-1))

       # Otherwise, delete all spaces until a non-space character is met
       else :
           for char in line [:delta] :
               if char != " " :
                   break
               cursor.deleteChar ()

# https://www.binpress.com/tutorial/building-a-text-editor-with-pyqt-part-2/145

# --------------------------------------------------------------------------

class Edit (IndentationTextEdit) :

   infoProperty = QTextFormat.UserProperty
   bookmarkProperty = QTextFormat.UserProperty + 1

   def __init__ (self, parent=None) :
       super (Edit, self).__init__ (parent)

       self.setLineWrapMode (QTextEdit.NoWrap) # otherwise selectLine is not exact

       self.setMouseTracking (True)

       if 1 :
          self.highlighter = Highlighter (self.document ()) # important: keep reference to highlighter

       if 1 :
          completerList = [ ]
          completerList.append ("alpha")
          completerList.append ("beta")
          completerList.append ("gamma")

          completer = QCompleter (completerList, self)
          self.setCompleter (completer, unfiltered = True)

   def selectLine (self, line) :
       cursor = self.textCursor ()
       cursor.movePosition (QTextCursor.Start)
       cursor.movePosition (QTextCursor.Down, QTextCursor.MoveAnchor, line-1)
       cursor.movePosition (QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
       self.setTextCursor (cursor)
       self.ensureCursorVisible ()

   def select (self, start_line, start_col, stop_line, stop_col) :
       cursor = self.textCursor ()
       cursor.movePosition (QTextCursor.Start)
       cursor.movePosition (QTextCursor.Down, QTextCursor.MoveAnchor, start_line-1)
       cursor.movePosition (QTextCursor.Right, QTextCursor.MoveAnchor, start_col-1)

       if start_line == stop_line :
          cursor.movePosition (QTextCursor.Right, QTextCursor.KeepAnchor, stop_col-start_col)
       else :
          cursor.movePosition (QTextCursor.Down, QTextCursor.KeepAnchor, stop_line-start_line)
          cursor.movePosition (QTextCursor.StartOfLine, QTextCursor.KeepAnchor)
          cursor.movePosition (QTextCursor.Right, QTextCursor.KeepAnchor, stop_col)

       self.setTextCursor (cursor)

   def event (self, e) :
       if False or e.type() == QEvent.ToolTip :
          cursor = self.cursorForPosition (e.pos ())
          cursor.select (QTextCursor.WordUnderCursor)
          if cursor.selectedText() != "" :
             fmt = cursor.charFormat ()
             if fmt.hasProperty (self.infoProperty) :
                tooltip = "info: " + fmt.stringProperty (self.infoProperty)
             elif fmt.hasProperty (QTextFormat.TextToolTip) :
                tooltip = "tooltip: " + fmt.toolTip()
             else :
                tooltip = cursor.selectedText ()
             QToolTip.showText (e.globalPos(), tooltip)
          else :
             QToolTip.hideText ()
          return True
       else :
          return super (Edit, self).event (e)

   def mouseMoveEvent (self, e) :
       if e.modifiers () == Qt.ControlModifier :
          cursor = self.cursorForPosition (e.pos ())
          cursor.select (QTextCursor.WordUnderCursor)
          if cursor.selectedText() != "" :
             fmt = cursor.charFormat ()
             if not fmt.fontUnderline () : # keep red underline
                fmt.setFontUnderline (True)
                fmt.setUnderlineColor (Qt.blue)
                cursor.setCharFormat (fmt)
       super (Edit, self).mouseMoveEvent (e)

   def mousePressEvent (self, e) :
       if e.modifiers () == Qt.ControlModifier :
          cursor = self.cursorForPosition (e.pos ())
          cursor.select (QTextCursor.WordUnderCursor)
          if cursor.selectedText() != "" :
             fmt = cursor.charFormat ()
             fmt.setFontUnderline (True)
             fmt.setUnderlineColor (Qt.red)
             cursor.setCharFormat (fmt)
       super (Edit, self).mousePressEvent (e)

   def contextMenuEvent (self, e) :
       menu = self.createStandardContextMenu ()
       menu.addSeparator ()
       menu.addAction ("My Menu Item")
       menu.exec_ (e.globalPos())

   def keyPressEvent (self, e) :
       mask = Qt.ShiftModifier | Qt.ControlModifier | Qt.AltModifier | Qt.MetaModifier
       mod = int (e.modifiers () & mask) # necessary for Ctrl Shift
       key = e.key ()
       # print "modifiers", format (mod, "08x")
       if mod == Qt.MetaModifier :
          if key >= Qt.Key_1 and key <= Qt.Key_5 :
             # Win 1 ... Win 5
             self.setBookmark (key - Qt.Key_0)
             return
       elif mod == Qt.ControlModifier :
          if key >= Qt.Key_1 and key <= Qt.Key_5 :
             # Ctrl 1 ... Ctrl 5
             # self.goToBookmark (key - Qt.Key_0)
             self.gotoBookmark (key - Qt.Key_0)
             return
       elif mod == Qt.ControlModifier + Qt.ShiftModifier :
          # Ctrl Shift 1 ... Ctrl Shift 5
          if key == Qt.Key_Exclam : # ord ('!')
             self.setBookmark (1)
             return
          if key == Qt.Key_At :
             self.setBookmark (2)
             return
          if key == Qt.Key_NumberSign :
             self.setBookmark (3)
             return
          if key == Qt.Key_Dollar :
             self.setBookmark (4)
             return
          if key == Qt.Key_Percent :
             self.setBookmark (5)
             return
       if (mod & Qt.AltModifier != 0) or (mod & Qt.MetaModifier != 0) :
          return
       super (Edit, self).keyPressEvent (e)

   bookmark_colors = [ QColor (Qt.yellow).lighter (160),
                       QColor (Qt.green).lighter (160),
                       QColor (Qt.blue).lighter (160),
                       QColor (Qt.red).lighter (160),
                       QColor (Qt.gray).lighter (140) ]

   def setBookmark (self, markType = 1) :
       line = self.textCursor ().blockNumber ()
       extraSelections = self.extraSelections ()

       found = False
       inx = 0
       for item in extraSelections :
           if not found :
              if item.format.hasProperty (self.bookmarkProperty) :
                 if item.format.intProperty (self.bookmarkProperty) == markType :
                    if item.cursor.blockNumber () == line :
                       del extraSelections [inx]
                       self.setExtraSelections (extraSelections)
                       found = True
           inx = inx + 1

       if not found :
          lineColor = self.bookmark_colors [markType-1]
          selection = QTextEdit.ExtraSelection ()
          selection.format.setBackground (lineColor)
          selection.format.setProperty (self.bookmarkProperty, markType)
          selection.format.setProperty (QTextFormat.FullWidthSelection, QVariant(True))
          selection.cursor = self.textCursor ()
          extraSelections.append (selection)
          self.setExtraSelections (extraSelections)

   def gotoBookmark (self, markType = 1) :
       line = self.textCursor ().blockNumber ()
       extraSelections = self.extraSelections ()
       found = False
       for item in extraSelections :
           if not found :
              if item.format.hasProperty (self.bookmarkProperty) :
                 if item.format.intProperty (self.bookmarkProperty) == markType :
                    if item.cursor.blockNumber () > line :
                       self.setTextCursor (item.cursor)
                       found = True
       for item in extraSelections :
          if not found :
              if item.format.hasProperty (self.bookmarkProperty) :
                 if item.format.intProperty (self.bookmarkProperty) == markType :
                    self.setTextCursor (item.cursor)
                    found = True

       #cursor = self.textCursor ()
       #stop = False
       #found = False
       #while not found and not stop :
          #if not cursor.movePosition (QTextCursor.Down) :
             #stop = True
          #if cursor.atEnd () :
             #stop = True
          #if cursor.charFormat ().hasProperty (self.bookmarkProperty) :
             #if cursor.charFormat ().intProperty (self.bookmarkProperty) == markType :
                #found = True
       #if found :
          #self.setTextCursor (cursor)
          #self.ensureCursorVisible ()


# http://www.qtcentre.org/threads/11791-QTextEdit-get-word-under-the-mouse-pointer

# --------------------------------------------------------------------------

class FindBox (QWidget) :

   def __init__ (self, parent = None) :
       super (FindBox, self).__init__ (parent)

       self.edit = None
       self.small = False

       self.horiz_layout = QHBoxLayout ()
       self.setLayout (self.horiz_layout)

       self.closeButton = QToolButton ()
       self.closeButton.setIcon (QIcon.fromTheme ("dialog-close")) # !?
       self.closeButton.setShortcut ("Esc")
       # self.closeButton.setMaximumWidth (20)
       # self.closeButton.setFlat (True)
       self.closeButton.clicked.connect (self.hide)
       self.horiz_layout.addWidget (self.closeButton)

       self.label = QLabel ()
       self.label.setText ("Find:")
       self.horiz_layout.addWidget (self.label)

       # self.line = QLineEdit ()
       self.line = QComboBox ()
       self.line.setEditable (True)
       # self.line.setFrame (False)
       # self.line.setDuplicatesEnabled (False)
       # self.line.setInsertPolicy (QComboBox.InsertAtBottom)
       self.horiz_layout.addWidget (self.line)

       self.wholeWords = QCheckBox ()
       self.wholeWords.setText ("Whole &words")
       self.horiz_layout.addWidget (self.wholeWords)

       self.matchCase = QCheckBox ()
       self.matchCase.setText ("Match &case")
       self.horiz_layout.addWidget (self.matchCase)

       self.nextButton = QPushButton ()
       self.nextButton.setText ("&Next")
       self.nextButton.clicked.connect (self.findNext)
       self.horiz_layout.addWidget (self.nextButton)

       self.prevButton = QPushButton ()
       self.prevButton.setText ("&Prev")
       self.prevButton.clicked.connect (self.findPrev)
       self.horiz_layout.addWidget (self.prevButton)

       self.horiz_layout.setStretch (2, 10)

       # self.line.returnPressed.connect (self.returnPressed)
       self.line.lineEdit().returnPressed.connect (self.returnPressed)

   def resizeEvent (self, e) :
       w = e.size().width()
       if w < 640 :
          if not self.small :
             self.wholeWords.setText ("&Words")
             self.matchCase.setText ("&Case")
             self.small = True
       else :
          if self.small :
             self.wholeWords.setText ("Whole &words")
             self.matchCase.setText ("Match &case")
             self.small = False
       super (FindBox, self).resizeEvent (e)

   def returnPressed (self) :
       self.findNext ()

   def open (self) :
       self.show ()
       self.line.setFocus ()
       self.line.lineEdit ().selectAll () # enable editing

   def findNext (self) :
       self.find (False)

   def findPrev (self) :
       self.find (True)

   def find (self, back) :
       # text = self.line.text ()
       text = self.line.currentText ()
       if text != "" :
          if self.line.findText (text) == -1 :
             self.line.addItem (text)

          opt = QTextDocument.FindFlags (0)
          if self.wholeWords.isChecked () :
             opt = opt | QTextDocument.FindWholeWords
          if self.matchCase.isChecked () :
             opt = opt | QTextDocument.FindCaseSensitively
          if back :
             opt = opt | QTextDocument.FindBackward

          ok = self.edit.find (text, opt)
          if not ok :
             if back :
                self.edit.moveCursor (QTextCursor.End)
             else :
                self.edit.moveCursor (QTextCursor.Start)
             self.edit.find (text, opt)

# --------------------------------------------------------------------------

class GoToLineBox (QWidget) :

   def __init__ (self, parent=None) :
       super (GoToLineBox, self).__init__ (parent)

       self.edit = None

       self.horiz_layout = QHBoxLayout ()
       self.setLayout (self.horiz_layout)

       self.closeButton = QToolButton ()
       self.closeButton.setIcon (QIcon.fromTheme ("dialog-close")) # !?
       self.closeButton.setShortcut ("Esc")
       self.closeButton.clicked.connect (self.hide)
       self.horiz_layout.addWidget (self.closeButton)

       self.label = QLabel ()
       self.label.setText ("Go to line:")
       self.horiz_layout.addWidget (self.label)

       self.line = QSpinBox ()
       self.horiz_layout.addWidget (self.line)

       self.button = QPushButton ()
       self.button.setText ("Go")
       self.horiz_layout.addWidget (self.button)

       self.horiz_layout.addStretch ()

       self.line.editingFinished.connect (self.editingFinished)

   def open (self) :
       cursor = self.edit.textCursor ()
       document = self.edit.document ()
       self.line.setMinimum (1)
       self.line.setMaximum (document.blockCount ())
       self.line.setValue (cursor.blockNumber () + 1) # line number
       self.line.setFocus ()
       self.line.selectAll () # enable editing
       self.show ()

   def editingFinished (self) :
       self.edit.selectLine (self.line.value ())

# --------------------------------------------------------------------------

class Editor (QWidget) :

   def __init__ (self, parent=None) :
       super (Editor, self).__init__ (parent)

       self.origFileName = ""

       self.layout = QVBoxLayout ()
       self.setLayout (self.layout)

       self.edit = Edit ()
       self.layout.addWidget (self.edit)

       self.findBox = FindBox ()
       self.findBox.edit = self.edit
       self.findBox.hide ()
       self.layout.addWidget (self.findBox)

       self.lineBox = GoToLineBox ()
       self.lineBox.edit = self.edit
       self.lineBox.hide ()
       self.layout.addWidget (self.lineBox)

       if 1 :
          font = QFont ("Luxi Mono", 14)
          font.setFixedPitch (True)
          self.edit.setFont (font)

   def indent (self) :
       self.edit.indent ()

   def unindent (self) :
       self.edit.unindent ()

   def selectLine (self, line) :
       self.edit.selectLine (line)

   def select (self, start_line, start_col, stop_line, stop_col) :
       self.edit.select (start_line, start_col, stop_line, stop_col)

   def setBookmark (self, markType = 1) :
       self.setBookmark (markType)

   def find (self) :
       self.lineBox.hide ()
       self.findBox.open ()

   def findNext (self) :
       self.findBox.findNext ()

   def findPrev (self) :
       self.findBox.findPrev ()

   def replace (self) :
       self.lineBox.hide ()
       self.findBox.open ()

   def goToLine (self) :
       self.findBox.hide ()
       self.lineBox.open ()

   def simpleReadFile (self, fileName) :
       f = open (fileName)
       text = f.read ()
       self.edit.setPlainText (text)

   def readFile (self, fileName) :
       self.origFileName = os.path.abspath (fileName)
       try :
          f = QFile (fileName)
          if not f.open (QIODevice.ReadOnly) :
             raise IOError (unicode (f.errorString()))

          stream = QTextStream (f)
          stream.setCodec ("UTF-8")
          text = stream.readAll ()

          self.edit.setPlainText (text)
          self.edit.document ().setModified (False)

          fi = QFileInfo (fileName)
          self.setWindowTitle (fi.fileName ())
       except IOError as e :
          QMessageBox.warning (self, "Read File Error", "Cannot open file %s: %s" % (fileName, e))
       except OSError as e :
          QMessageBox.warning (self, "Read File Error", "Cannot open file %s: %s" % (fileName, e))

   def writeFile (self, fileName) :
       self.origFileName = os.path.abspath (fileName)
       fh = None
       try :
           fh = QFile (fileName)
           if not fh.open (QIODevice.WriteOnly) :
              raise IOError (unicode (fh.errorString ()))
           stream = QTextStream (fh)
           stream.setCodec ("UTF-8")
           stream << self.edit.toPlainText ()
           self.edit.document ().setModified (False)
       except IOError as e :
           QMessageBox.warning (self, "Write File Error", "Failed to save %s: %s" % (self.fileName, e))
       except OSError as e :
           QMessageBox.warning (self, "Write File Error", "Failed to save %s: %s" % (self.fileName, e))
       finally :
           if fh is not None :
              fh.close ()

   def isModified (self) :
       return self.edit.document ().isModified ()

   def openFile (self) :
       fileName = str (QFileDialog.getOpenFileName (self, "Open File"))
       if fileName != "" :
          self.readFile (fileName)

   def saveFile (self) :
       self.writeFile (self.origFileName)

   def saveFileAs (self) :
       fileName = str (QFileDialog.getSaveFileName (self, "Save File As", self.origFileName))
       if fileName != "" :
          self.writeFile (fileName)

# --------------------------------------------------------------------------

class EditWin (QMainWindow) :

    def __init__ (self, parent=None) :
        super (EditWin, self).__init__ ( parent)

        self.editor = Editor ()
        self.setCentralWidget (self.editor)

        # File menu

        fileMenu = self.menuBar().addMenu ("&File")

        act = QAction ("&Open...", self)
        act.triggered.connect (self.editor.openFile)
        fileMenu.addAction (act)

        act = QAction ("&Quit", self)
        act.triggered.connect (self.close)
        fileMenu.addAction (act)

        # Edit Menu

        editMenu = self.menuBar().addMenu ("&Edit")

        act = QAction ("&Indent", self)
        act.setShortcut ("Ctrl+I")
        act.setToolTip ("Indent the current line or selection")
        act.setStatusTip (act.toolTip())
        act.triggered.connect (self.editor.indent)
        editMenu.addAction (act)

        act = QAction ("&Unindent", self)
        act.setShortcut ("Ctrl+U")
        act.setToolTip ("Unindent the current line or selection")
        act.setStatusTip (act.toolTip())
        act.triggered.connect (self.editor.unindent)
        editMenu.addAction (act)

        editMenu.addSeparator ()

        act = QAction ("&Find", self)
        act.setShortcut ("Ctrl+F")
        act.triggered.connect (self.editor.find)
        editMenu.addAction (act)

        act = QAction ("&Replace", self)
        act.setShortcut ("Ctrl+R")
        act.triggered.connect (self.editor.replace)
        editMenu.addAction (act)

        act = QAction ("Find &Next", self)
        act.setShortcut ("F3")
        act.triggered.connect (self.editor.findBox.findNext)
        editMenu.addAction (act)

        act = QAction ("Find &Previous", self)
        act.setShortcut ("Shift+F3")
        act.triggered.connect (self.editor.findBox.findPrev)
        editMenu.addAction (act)

        editMenu.addSeparator ()

        act = QAction ("&Go to Line", self)
        act.setShortcut ("Ctrl+G")
        act.triggered.connect (self.editor.goToLine)
        editMenu.addAction (act)

        act = QAction ("Set &Bookmark", self)
        act.setShortcut ("Ctrl+B")
        act.triggered.connect (self.editor.setBookmark)
        editMenu.addAction (act)

# --------------------------------------------------------------------------

if __name__ == '__main__' :
   app = QApplication (sys.argv)
   win = EditWin ()
   win.show ()
   win.editor.readFile ("examples/example.cc")
   win.editor.select (6, 5, 6, 8)
   app.exec_ ()

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
