#!/usr/bin/env python

from __future__ import print_function

import sys, os

from util import use_qt5
if use_qt5 :
   from PyQt5.QtCore import *
   from PyQt5.QtGui import *
   from PyQt5.QtWidgets import *
else :
   from PyQt4.QtCore import *
   from PyQt4.QtGui import *

from edit import Editor
from area import GraphicsWin, GraphicsWithPalette
from tree import Tree
from treedir import TreeDir
from prop import Property
from treeprop import TreeProperty
from grep import GrepWin
from output import Output
from tools import refreshToolsAndHelp

from simple import SimplePlugin

sys.path.insert (1, os.path.join (os.getcwd (), "../grm")) # directory with parser

from gram_plugin import GrammarPlugin

if sys.version_info >= (3,) :
   import importlib
   def reload (name) :
      importlib.reload (name)

# --------------------------------------------------------------------------

class DetailWindow (QMainWindow):
   def __init__ (self, parent = None) :
       super (DetailWindow, self).__init__ (parent)

       self.items = { }

       self.tabWidget = QTabWidget ()
       self.tabWidget.setMovable (True)
       self.setCentralWidget (self.tabWidget)

   def addItem (self, fileName) :
       if fileName not in self.items :
          w = Editor ()
          self.items [fileName] = w
          name = os.path.basename (fileName)
          self.tabWidget.addTab (w, name)
       w = self.items [fileName]
       self.tabWidget.setCurrentWidget (w)
       return w

# --------------------------------------------------------------------------

class ViewWindow (QMainWindow):

   def __init__ (self, parent = None) :
       super (ViewWindow, self).__init__ (parent)

       self.icon_path = [ ]
       self.icon_cache = { }

       self.color_cache = { }

       self.editors = { }

       self.settings = QSettings ("view.ini", QSettings.IniFormat)
       self.session = QSettings ("view.cfg", QSettings.IniFormat)

       self.other_win = DetailWindow (self)
       self.zoom_win = DetailWindow ()

       self.window_list = [ self, self.other_win, self.zoom_win ]

       self.toolbar = self.addToolBar ("Main")
       self.status = self.statusBar ()

       # self.statusLabel = QLabel ()
       # self.status.addWidget (self.statusLabel)

       # left

       self.leftTabs = QTabWidget ()
       self.leftTabs.setMovable (True)

       self.tree = Tree ()
       self.tree.win = self
       self.leftTabs.addTab (self.tree, "Tree")

       self.files = TreeDir ()
       self.files.win = self
       self.leftTabs.addTab (self.files, "Files")

       self.grep = GrepWin ()
       self.grep.win = self
       self.leftTabs.addTab (self.grep, "Grep")

       # center

       self.tabWidget = QTabWidget ()
       self.tabWidget.setMovable (True)

       self.design = GraphicsWithPalette ()
       self.tabWidget.addTab (self.design, "Design")

       # right

       self.rightTabs = QTabWidget ()
       self.rightTabs.setMovable (True)

       self.prop = Property ()
       self.rightTabs.addTab (self.prop, "Properties")

       self.treeProp = TreeProperty ()
       self.rightTabs.addTab (self.treeProp, "Tree Properties")

       # bottom

       self.info = Output ()
       self.info.win = self

       # layout

       self.hsplitter = QSplitter ()
       self.hsplitter.addWidget (self.leftTabs)
       self.hsplitter.addWidget (self.tabWidget)
       self.hsplitter.addWidget (self.rightTabs)

       self.vsplitter = QSplitter ()
       self.vsplitter.setOrientation (Qt.Vertical)
       self.vsplitter.addWidget (self.hsplitter)
       self.vsplitter.addWidget (self.info)
       self.vsplitter.setStretchFactor (0, 3)
       self.vsplitter.setStretchFactor (1, 1)

       self.setCentralWidget (self.vsplitter)

       # file menu

       self.fileMenu = self.menuBar().addMenu ("&File")
       self.fileMenu.aboutToShow.connect (self.onShowFileMenu)
       menu = self.fileMenu

       act = QAction ("&Open...", self)
       act.setShortcut ("Ctrl+O")
       act.triggered.connect (self.openFile)
       menu.addAction (act)

       act = QAction ("&Save", self)
       act.setShortcut ("Ctrl+S")
       act.triggered.connect (self.saveFile)
       menu.addAction (act)
       self.saveAction = act

       act = QAction ("Save As...", self)
       act.triggered.connect (self.saveFileAs)
       menu.addAction (act)
       self.saveAsAction = act

       act = QAction ("&Close", self)
       act.setShortcut ("Ctrl+W")
       act.triggered.connect (self.closeFile)
       menu.addAction (act)
       self.closeAction = act

       menu.addSeparator ()

       act = QAction ("&Quit", self)
       act.setShortcut ("Ctrl+Q")
       act.triggered.connect (self.close)
       menu.addAction (act)

       # edit menu

       self.editMenu = self.menuBar().addMenu ("&Edit")
       self.editMenu.aboutToShow.connect (self.onShowEditMenu)
       menu = self.editMenu

       act = QAction ("&Indent", self)
       act.setShortcut ("Ctrl+I")
       act.setStatusTip ("Indent the current line or selection")
       act.triggered.connect (self.indent)
       menu.addAction (act)

       act = QAction ("&Unindent", self)
       act.setShortcut ("Ctrl+U")
       act.setStatusTip ("Unindent the current line or selection")
       act.triggered.connect (self.unindent)
       menu.addAction (act)

       menu.addSeparator ()

       act = QAction ("&Find", self)
       act.setShortcut ("Ctrl+F")
       act.triggered.connect (self.find)
       menu.addAction (act)

       act = QAction ("&Replace", self)
       act.setShortcut ("Ctrl+R")
       act.triggered.connect (self.replace)
       menu.addAction (act)

       act = QAction ("Find &Next", self)
       act.setShortcut ("F3")
       act.triggered.connect (self.findNext)
       menu.addAction (act)

       act = QAction ("Find &Previous", self)
       act.setShortcut ("Shift+F3")
       act.triggered.connect (self.findPrev)
       menu.addAction (act)

       menu.addSeparator ()

       act = QAction ("&Go to Line", self)
       act.setShortcut ("Ctrl+G")
       act.triggered.connect (self.goToLine)
       menu.addAction (act)

       act = QAction ("Set &Bookmark", self)
       act.setShortcut ("Ctrl+B")
       act.triggered.connect (self.setBookmark)
       menu.addAction (act)

       # view menu

       self.viewMenu = self.menuBar().addMenu ("&View")
       # self.viewMenu.aboutToShow.connect (self.onShowViewMenu)
       menu = self.viewMenu

       # project menu

       self.runMenu = self.menuBar().addMenu ("&Project")
       menu = self.runMenu

       act = QAction ("&Grep", self)
       act.setShortcut ("F2")
       act.triggered.connect (self.findInFiles)
       menu.addAction (act)

       act = QAction ("Grep in output", self)
       act.triggered.connect (self.grepInOutput)
       menu.addAction (act)

       menu.addSeparator ()

       act = QAction ("Configure", self)
       act.triggered.connect (self.configure)
       menu.addAction (act)

       act = QAction ("&Build", self)
       act.setShortcut ("F7")
       act.triggered.connect (self.build)
       menu.addAction (act)

       act = QAction ("&Make", self)
       act.setShortcut ("F8")
       act.triggered.connect (self.make)
       menu.addAction (act)

       act = QAction ("&Run", self)
       act.triggered.connect (self.run)
       menu.addAction (act)

       act = QAction ("&Debug", self)
       act.triggered.connect (self.debug)
       menu.addAction (act)

       act = QAction ("Install", self)
       act.triggered.connect (self.install)
       menu.addAction (act)

       act = QAction ("Clean", self)
       act.triggered.connect (self.clean)
       menu.addAction (act)

       menu.addSeparator ()

       act = QAction ("&Stop", self)
       act.setShortcut ("Ctrl+Esc")
       act.triggered.connect (self.info.stop)
       menu.addAction (act)
       self.info.stopAction = act

       act = QAction ("&Clear output", self)
       act.triggered.connect (self.info.clearOutput)
       menu.addAction (act)

       menu.addSeparator ()

       act = QAction ("Next mark", self)
       act.setShortcut ("F4")
       act.setStatusTip ("Jump to next output mark")
       act.triggered.connect (self.info.jumpToNextMark)
       menu.addAction (act)

       act = QAction ("Previous mark", self)
       act.setShortcut ("Shift+F4")
       act.setStatusTip ("Jump to previous output mark")
       act.triggered.connect (self.info.jumpToPrevMark)
       menu.addAction (act)

       menu.addSeparator ()

       act = QAction ("Re-read INI settings", self)
       act.setStatusTip ("Re-read settings from view.ini file")
       act.triggered.connect (self.refreshIni)
       menu.addAction (act)

       act = QAction ("Re-read session", self)
       act.setStatusTip ("Re-read session from view.cfg file")
       act.triggered.connect (self.reloadSession)
       menu.addAction (act)

       act = QAction ("Re-load modules", self)
       act.setStatusTip ("Re-load some python modules and editor files")
       act.triggered.connect (self.reloadModules)
       menu.addAction (act)

       # grammar menu

       GrammarPlugin (self)

       # simple menu

       SimplePlugin (self)

       # window menu

       self.windowMenu = self.menuBar().addMenu ("&Window")
       self.windowMenu.aboutToShow.connect (self.onShowWindowMenu)
       menu = self.windowMenu

       self.showTabActiom (menu, "Tree",       self.leftTabs,  self.tree,  "Meta+T")
       self.showTabActiom (menu, "Files",      self.leftTabs,  self.files, "Meta+F")
       self.showTabActiom (menu, "Grep",       self.leftTabs,  self.grep,  "Meta+G")
       self.showTabActiom (menu, "Properties", self.rightTabs, self.prop,  "Meta+P")

       menu.addSeparator ()

       self.leftAction  = self.showDockAction (menu, "&Left",  self.leftTabs,  "Ctrl+,")
       self.rightAction = self.showDockAction (menu, "&Right", self.rightTabs, "Ctrl+.")
       self.infoAction  = self.showDockAction (menu, "&Info",  self.info,      "Ctrl+/")

       menu.addSeparator ()

       act = QAction ("Other window", self)
       act.setShortcut ("Meta+O")
       act.setStatusTip ("Open text in another window")
       act.triggered.connect (self.showOther)
       menu.addAction (act)
       self.otherAction = act

       act = QAction ("&Zoom window", self)
       act.setShortcut ("Meta+Z")
       act.setStatusTip ("Open text in zoom window")
       act.triggered.connect (self.showZoom)
       menu.addAction (act)
       self.zoomAction = act

       menu.addSeparator ()

       act = QAction ("Main window", self)
       act.setShortcut ("Meta+Home")
       act.setShortcutContext (Qt.ApplicationShortcut)
       act.setStatusTip ("Select main window")
       act.triggered.connect (self.selectMainWindow)
       menu.addAction (act)

       act = QAction ("Left window", self)
       act.setShortcut ("Meta+Left")
       act.setShortcutContext (Qt.ApplicationShortcut)
       act.setStatusTip ("Select left window")
       act.triggered.connect (self.selectLeftWindow)
       menu.addAction (act)

       act = QAction ("Right window", self)
       act.setShortcut ("Meta+Right")
       act.setShortcutContext (Qt.ApplicationShortcut)
       act.setStatusTip ("Select right window")
       act.triggered.connect (self.selectRightWindow)
       menu.addAction (act)

       act = QAction ("Above window", self)
       act.setShortcut ("Meta+Up")
       act.setShortcutContext (Qt.ApplicationShortcut)
       act.setStatusTip ("Select above window")
       act.triggered.connect (self.selectAboveWindow)
       menu.addAction (act)

       act = QAction ("Below window", self)
       act.setShortcut ("Meta+Down")
       act.setShortcutContext (Qt.ApplicationShortcut)
       act.setStatusTip ("Select below window")
       act.triggered.connect (self.selectBelowWindow)
       menu.addAction (act)

       # tools menu

       self.toolsMenu = self.menuBar().addMenu ("&Tools")
       menu = self.toolsMenu

       # help menu

       self.helpMenu = self.menuBar().addMenu ("&Help")
       menu = self.helpMenu

       refreshToolsAndHelp (self)

       # keyboard shortcuts

       for inx in range (10) :
          self.showPageShortcut (inx)

   # Window menu

   def onShowWindowMenu (self) :
       self.leftAction.setChecked  (self.leftTabs.isVisible ())
       self.rightAction.setChecked (self.rightTabs.isVisible ())
       self.infoAction.setChecked  (self.info.isVisible ())

       is_edit = self.getEditor() != None
       self.otherAction.setEnabled (is_edit)
       self.zoomAction.setEnabled (is_edit)

   def showTabActiom (self, menu, text, tabWidget, widget, shortcut) :
       act = QAction (text, self)
       act.setShortcut (shortcut)
       act.setShortcutContext (Qt.ApplicationShortcut)
       act.triggered.connect (lambda: self.showTabAndWin (tabWidget, widget))
       menu.addAction (act)
       return act

   def showTabAndWin (self, tabWidget, widget) :
       tabWidget.setVisible (True)
       tabWidget.setCurrentWidget (widget)

   def showDockAction (self, menu, text, widget, shortcut) :
       act = QAction (text, self)
       act.setCheckable (True)
       act.setChecked (True)
       act.setShortcut (shortcut)
       act.setShortcutContext (Qt.ApplicationShortcut)
       act.triggered.connect (widget.setVisible)
       menu.addAction (act)
       return act

   def showPageShortcut (self, inx) :
       act = QShortcut (self)
       act.setKey ("Alt+" + str (inx))
       act.setContext (Qt.ApplicationShortcut)
       act.activated.connect (lambda: self.setTab (inx))

   def selectWindow (self, compare)  :
       # print ("SELECT WINDOW")

       win_list = [ w for w in self.window_list if w.isVisible () ]
       win_list = sorted (win_list, compare)

       active = QApplication.activeWindow ()
       # for w in win_list :
       #     print ("WINDOW", w.windowTitle(), w.__class__, w.x (), w.y (), w == active)

       target = None
       cnt = len (win_list)
       inx = -1
       if active in win_list :
          inx = win_list.index (active)

       if cnt > 0 and inx >= 0 :
          if inx > 0 :
             target = win_list [inx-1]
          else :
             target = win_list [-1]

       if target != None and target != active :
          # QApplication.setActiveWindow (target)
          target.activateWindow ()
          target.raise_ ()

   def selectMainWindow (self) :
       self.activateWindow ()
       self.raise_ ()

   def selectLeftWindow (self) :
       self.selectWindow (lambda a, b : a.x () < b.x ())

   def selectRightWindow (self) :
       self.selectWindow (lambda a, b : a.x () > b.x ())

   def selectAboveWindow (self) :
       self.selectWindow (lambda a, b : a.y () < b.y ())

   def selectBelowWindow (self) :
       self.selectWindow (lambda a, b : a.y () < b.y ())

   # Detail windows

   def showOther (self) :
       editor = self.getEditor ()
       if editor != None :
          fileName = editor.origFileName
          w = self.other_win.addItem (fileName)
          e = w.edit
          e.setDocument (editor.edit.document ())
          self.other_win.show ()

   def showZoom (self) :
       editor = self.getEditor ()
       if editor != None :
          fileName = editor.origFileName
          w = self.zoom_win.addItem (fileName)
          e = w.edit
          e.setReadOnly (True)
          e.setDocument (editor.edit.document ().clone())
          e.setStyleSheet ("QTextEdit {font-size: 24px}")
          # font = QFont ("Luxi Mono", 24)
          # font.setFixedPitch (True)
          # e.setFont (font)
          e.setLineWrapMode (QTextEdit.NoWrap)
          self.zoom_win.show ()

   # Store session

   def storeSession (self) :
       self.session.clear ()
       inx = 0
       for key in self.editors :
           editor = self.editors [key]
           inx = inx + 1
           self.session.beginGroup ("edit-" + str (inx))
           self.session.remove ("")
           self.session.setValue ("fileName", editor.origFileName)
           cursor = editor.edit.textCursor ()
           self.session.setValue ("line", cursor.blockNumber ())
           self.session.setValue ("column", cursor.columnNumber ())

           self.session.beginWriteArray ("selections")
           extra = editor.edit.extraSelections ()
           n = 0
           for item in extra :
               if item.format.hasProperty (editor.edit.bookmarkProperty) :
                  self.session.setArrayIndex (n)
                  cursor = item.cursor
                  self.session.setValue ("line", cursor.blockNumber () + 1)
                  self.session.setValue ("column", cursor.columnNumber ())
                  index = item.format.intProperty (editor.edit.bookmarkProperty)
                  self.session.setValue ("index", index)
                  n = n + 1
           self.session.endArray ()
           self.session.endGroup ()

   def reloadSession (self) :
       groups = win.session.childGroups ()
       for group in groups :
           group_name = str (group)
           if group_name.startswith ("edit") :
              self.session.beginGroup (group_name)
              fileName = self.session.value ("fileName", "", str)
              line = self.session.value ("line", 0, int)
              column = self.session.value ("column", 0, int)
              editor = self.loadFile (fileName) # , line, column)

              selections = editor.edit.extraSelections ()
              count = self.session.beginReadArray ("selections")
              for inx in range (count) :
                  self.session.setArrayIndex (inx)
                  line = self.session.value ("line", 0, int)
                  column = self.session.value ("column", 0, int)

                  selection = QTextEdit.ExtraSelection ()
                  selection.cursor = editor.edit.textCursor () # important
                  selection.cursor.movePosition (QTextCursor.Start)
                  selection.cursor.movePosition (QTextCursor.NextBlock, QTextCursor.MoveAnchor, line-1)
                  index = self.session.value ("index", 0, int)
                  if (index == 0) :
                     index = 1
                  selection.format.setBackground (editor.edit.bookmark_colors [index-1])
                  selection.format.setProperty (editor.edit.bookmarkProperty, index)
                  selection.format.setProperty (QTextFormat.FullWidthSelection, QVariant(True))

                  selections.append (selection)

              editor.edit.setExtraSelections (selections)
              self.session.endArray ()
              self.session.endGroup ()

   def checkSave (self, editor) :
       if not editor.isModified () :
          return True
       else :
          dialog = QMessageBox (self)
          dialog.setIcon (QMessageBox.Warning)
          dialog.setText ("The " + editor.origFileName + " has been modified")
          dialog.setInformativeText ("Do you want to save your changes?")
          dialog.setStandardButtons (QMessageBox.Save | QMessageBox.Cancel | QMessageBox.Discard)
          dialog.setDefaultButton (QMessageBox.Save)
          answer = dialog.exec_ ()
          if answer == QMessageBox.Save :
             editor.writeFile (editor.origFileName)
             return True
          elif answer == QMessageBox.Discard :
             return True
          else :
             return False

   def closeEvent (self, e) :
       self.storeSession ()
       for key in self.editors :
           editor = self.editors [key]
           if not self.checkSave (editor) :
              e.ignore ()
              return
       e.accept ()
       self.other_win.close ()
       self.zoom_win.close ()

   # Settings

   def refreshIni (self) :
       self.settings.sync ()
       refreshToolsAndHelp (self)

   def stringValue (self, name, init = "") :
       result = self.settings.value (name, init)
       if use_qt5 :
          return str (result)
       else :
          return str (result.toString ())

   # Icons

   def addIconDir (self, path):
       path = os.path.expanduser (path)
       path = os.path.abspath (path)
       self.icon_path.append (path)

   def findIcon (self, name):
       if name in self.icon_cache :
          icon = self.icon_cache [name]
       else :
          icon = None
          for path in self.icon_path :
              file_name = os.path.join (path, name + ".png")
              if os.path.isfile (file_name) :
                 icon = QIcon (os.path.join (path, name + ".png"))
                 break
          if icon == None :
             icon = QIcon.fromTheme (name)
          if icon != None :
             self.icon_cache [name] = icon
       return icon

   def defineIcon (self, alias, name):
       if alias not in self.icon_cache : # if alias is not already defined
          icon = self.findIcon (name)
          if icon != None :
             self.icon_cache [alias] = icon # store icon under new name

   # Colors

   def findColor (self, name):
       if name in self.color_cache :
          return self.color_cache [name]
       else :
          color = QColor (name)
          self.color_cache [name] = color
          return color

   def defineColor (self, alias, color):
       self.color_cache [alias] = self.findColor (color)

   # Properties

   def showProperties (self, data) :
       self.prop.showProperties (data)

   # Editors

   def getEditor (self) :
       e = self.tabWidget.currentWidget ()
       if isinstance (e, Editor) :
          return e
       else :
          return None

   def hasExtension (self, editor, extension) :
       if editor == None :
          return False
       else :
          name, ext = os.path.splitext (editor.origFileName)
          return ext == extension

   # Files

   def loadFile (self, fileName, line = 0, column = 0, reload = False) :
       fileName = str (fileName) # optionally convert QStrint to string
       if fileName == "" or not os.path.isfile (fileName) :
          return None
       else :
          fileName = os.path.abspath (fileName)
          if fileName in self.editors :
             editor = self.editors [fileName]
             if reload :
                editor.readFile (fileName)
          else :
             editor = Editor ()
             self.editors [fileName] = editor
             editor.readFile (fileName)
             name = os.path.basename (fileName)
             self.tabWidget.addTab (editor, name)
          self.tabWidget.setCurrentWidget (editor)
          if line != 0 :
             editor.selectLine (line)
          return editor

   def reloadFile (self, fileName) :
       self.loadFile (fileName, reload = True)

   # File menu

   def onShowFileMenu (self) :
       is_edit = self.getEditor() != None
       self.saveAction.setEnabled (is_edit)
       self.saveAsAction.setEnabled (is_edit)
       self.closeAction.setEnabled (is_edit)

   def openFile (self) :
       fileName = str (QFileDialog.getOpenFileName (self, "Open File"))
       if fileName != "" :
          self.loadFile (fileName)

   def saveFile (self) :
       editor = self.getEditor ()
       if editor != None :
          # editor.saveFile ()
          editor.writeFile (editor.origFileName)

   def saveFileAs (self) :
       editor = self.getEditor ()
       if editor != None :
          # editor.saveFileAs ()
          oldFileName = editor.origFileName
          fileName = str (QFileDialog.getSaveFileName (self, "Save File As", editor.origFileName))
          if fileName != "" :
             del self.editors [editor.origFileName]
             editor.writeFile (fileName)
             self.editors [editor.origFileName] = editor

   def closeFile (self) :
       editor = self.getEditor ()
       if editor != None :
          if self.checkSave (editor) :
             del self.editors [editor.origFileName]
             inx = self.tabWidget.indexOf (editor)
             if inx >= 0 :
                self.tabWidget.removeTab (inx)

   # Edit menu

   def onShowEditMenu (self) :
       is_edit = self.getEditor() != None
       for act in self.editMenu.actions () :
          act.setEnabled (is_edit)

   def indent (self) :
       e = self.tabWidget.currentWidget ()
       if isinstance (e, Editor) :
          e.indent ()

   def unindent (self) :
       e = self.getEditor ()
       if e != None :
          e.unindent ()

   def find (self) :
       e = self.getEditor ()
       if e != None :
          e.find ()

   def findNext (self) :
       e = self.getEditor ()
       if e != None :
          e.findNext ()

   def findPrev (self) :
       e = self.getEditor ()
       if e != None :
          e.findPrev ()

   def replace (self) :
       e = self.getEditor ()
       if e != None :
          e.replace ()

   def goToLine (self) :
       e = self.getEditor ()
       if e != None :
          e.goToLine ()

   def setBookmark (self) :
       e = self.getEditor ()
       if e != None :
          e.setBookmark (markType)

   def setTab (self, inx) :
       # print ("TAB", inx)
       cnt = self.tabWidget.count ()
       if inx == 0 :
          inx = cnt # 0 ... last tabs
       if inx >= 1 and inx <= cnt :
          self.tabWidget.setCurrentIndex (inx-1)

   # View menu

   # def onShowViewMenu (self) :

   # Project menu

   def findInFiles (self) :
       self.leftTabs.setCurrentWidget (self.grep)
       self.grep.dialog ()

   def grepInOutput (self) :
       text = ""
       text, ok = QInputDialog.getText (self, "Grep", "Find in files", QLineEdit.Normal, text)
       if ok and text != "" :
          self.info.grep (text)

   def configure (self) :
       self.info.run ("configure")

   def build (self) :
       self.info.run ("build")

   def make (self) :
       self.info.run ("make")

   def run (self) :
       self.info.run ("run")

   def debug (self) :
       self.info.run ("debuf")

   def clean (self) :
       self.info.run ("clean")

   def install (self) :
       self.info.run ("install")

   # utilities

   def showHtml (self, fileName) :
       os.system ("firefox" + " " + 1 * "--new-window " + fileName + " &")

   def showPdf (self, fileName) :
       os.system ("atril" + " " + fileName + " &")

   def showPreview (self, fileName) :
       os.system ("caja" + " " + fileName + " &")

   # reload modules

   def reloadModules (self) :

       reload (sys.modules ['simple'])
       reload (sys.modules ['gram_plugin'])

       reload (sys.modules ['output'])

       self.reloadFile ("../grm/pas.g")
       self.reloadFile ("../grm/tkit.g")

# --------------------------------------------------------------------------

app = QApplication (sys.argv)
# app.setStyleSheet ("QWidget {font-size: 16px} Edit {font-family: \"Luxi Mono\" ; font-size: 16px} QDetailWindow {font-size: 24px} ")
app.setStyleSheet ("QWidget {font-size: 16px}")
QIcon.setThemeName ("oxygen") # /usr/share/icons/oxygen

win = ViewWindow ()
win.show ()

win.addIconDir ("~/temp/monodevelop-2.8.8.4/src/core/MonoDevelop.Ide/icons")
# win.addIconDir ("~/temp/monodevelop-4.0.14/src/core/MonoDevelop.Ide/icons")
# win.addIconDir ("~/temp/monodevelop-5.7/src/core/MonoDevelop.Ide/icons/light")
# win.addIconDir ("~/temp/monodevelop-5.9/src/core/MonoDevelop.Ide/icons")

win.defineIcon ("namespace", "element-namespace-16")
win.defineIcon ("class",     "element-class-16")
win.defineIcon ("function",  "element-method-16")
win.defineIcon ("variable",  "element-property-16")

win.defineColor ("namespace", "green")
win.defineColor ("class",     "blue")
win.defineColor ("function",  "brown")
win.defineColor ("variable",  "orange")

# win.info.redirectOutput ()
# print ("loading files ...")

win.loadFile ("examples/example.cc")
win.loadFile ("examples/example.sim")
win.loadFile ("../grm/pas.g")
win.loadFile ("../grm/tkit.g")

# win.reloadSession ()

app.exec_ ()

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
